@@ -69,6 +69,7 @@
 
 #include <gst/mpegts/mpegts.h>
 #include <gst/base/gstbytewriter.h>
+#include <gst/mpegtsdemux/gstmpegdesc.h>
 
 #include "tsmuxcommon.h"
 #include "tsmuxstream.h"
@@ -123,6 +124,7 @@ tsmux_stream_new (guint16 pid, guint stream_type, guint stream_number)
   stream->state = TSMUX_STREAM_STATE_HEADER;
   stream->pi.pid = pid;
   stream->stream_type = stream_type;
+  stream->internal_stream_type = stream_type;
 
   stream->pes_payload_size = 0;
   stream->cur_pes_payload_size = 0;
@@ -135,6 +137,7 @@ tsmux_stream_new (guint16 pid, guint stream_type, guint stream_number)
     case TSMUX_ST_VIDEO_MPEG4:
     case TSMUX_ST_VIDEO_H264:
     case TSMUX_ST_VIDEO_HEVC:
+    case TSMUX_ST_VIDEO_VVC:
       if (stream_number > 0xF) {
         GST_WARNING
             ("video stream number %d is greater than 0xF. Setting to 0.",
@@ -143,13 +146,24 @@ tsmux_stream_new (guint16 pid, guint stream_type, guint stream_number)
       }
       stream->id = 0xE0 | stream_number;
       stream->pi.flags |= TSMUX_PACKET_FLAG_PES_FULL_HEADER;
-      stream->is_video_stream = TRUE;
+      stream->gst_stream_type = GST_STREAM_TYPE_VIDEO;
       supports_user_specified_stream_number = TRUE;
       break;
     case TSMUX_ST_VIDEO_JP2K:
       stream->id = 0xBD;
       stream->pi.flags |= TSMUX_PACKET_FLAG_PES_FULL_HEADER;
-      stream->is_video_stream = TRUE;
+      stream->gst_stream_type = GST_STREAM_TYPE_VIDEO;
+      break;
+    case TSMUX_ST_VIDEO_JPEG_XS:
+      stream->id = 0xBD;        /* Private Stream 1 */
+      stream->pi.flags |= TSMUX_PACKET_FLAG_PES_FULL_HEADER;
+      stream->gst_stream_type = GST_STREAM_TYPE_VIDEO;
+      break;
+    case TSMUX_ST_PS_VP9:
+      stream->id = 0xBD;        /* Private Stream 1 */
+      stream->pi.flags |= TSMUX_PACKET_FLAG_PES_FULL_HEADER;
+      stream->stream_type = TSMUX_ST_PRIVATE_DATA;
+      stream->gst_stream_type = GST_STREAM_TYPE_VIDEO;
       break;
     case TSMUX_ST_AUDIO_AAC:
     case TSMUX_ST_AUDIO_MPEG1:
@@ -160,7 +174,7 @@ tsmux_stream_new (guint16 pid, guint stream_type, guint stream_number)
             stream_number);
         stream_number = 0;
       }
-      stream->is_audio = TRUE;
+      stream->gst_stream_type = GST_STREAM_TYPE_AUDIO;
       stream->id = 0xC0 | stream_number;
       stream->pi.flags |= TSMUX_PACKET_FLAG_PES_FULL_HEADER;
       supports_user_specified_stream_number = TRUE;
@@ -174,18 +188,18 @@ tsmux_stream_new (guint16 pid, guint stream_type, guint stream_number)
       switch (stream_type) {
         case TSMUX_ST_VIDEO_DIRAC:
           stream->id_extended = 0x60;
-          stream->is_video_stream = TRUE;
+          stream->gst_stream_type = GST_STREAM_TYPE_VIDEO;
           break;
         case TSMUX_ST_PS_AUDIO_LPCM:
-          stream->is_audio = TRUE;
+          stream->gst_stream_type = GST_STREAM_TYPE_AUDIO;
           stream->id_extended = 0x80;
           break;
         case TSMUX_ST_PS_AUDIO_AC3:
-          stream->is_audio = TRUE;
+          stream->gst_stream_type = GST_STREAM_TYPE_AUDIO;
           stream->id_extended = 0x71;
           break;
         case TSMUX_ST_PS_AUDIO_DTS:
-          stream->is_audio = TRUE;
+          stream->gst_stream_type = GST_STREAM_TYPE_AUDIO;
           stream->id_extended = 0x82;
           break;
         default:
@@ -198,7 +212,12 @@ tsmux_stream_new (guint16 pid, guint stream_type, guint stream_number)
     case TSMUX_ST_PS_TELETEXT:
       /* needs fixes PES header length */
       stream->pi.pes_header_length = 36;
-      /* fall through */
+      stream->id = 0xBD;
+      stream->stream_type = TSMUX_ST_PRIVATE_DATA;
+      stream->pi.flags |=
+          TSMUX_PACKET_FLAG_PES_FULL_HEADER |
+          TSMUX_PACKET_FLAG_PES_DATA_ALIGNMENT;
+      break;
     case TSMUX_ST_PS_DVB_SUBPICTURE:
       /* private stream 1 */
       stream->id = 0xBD;
@@ -213,17 +232,43 @@ tsmux_stream_new (guint16 pid, guint stream_type, guint stream_number)
       /* FIXME: assign sequential extended IDs? */
       stream->id = 0xBD;
       stream->stream_type = TSMUX_ST_PRIVATE_DATA;
-      stream->is_meta = TRUE;
       stream->pi.flags |=
           TSMUX_PACKET_FLAG_PES_FULL_HEADER |
           TSMUX_PACKET_FLAG_PES_DATA_ALIGNMENT;
       break;
+    case TSMUX_ST_PS_ST_2038:
+      /* FIXME: assign sequential extended IDs? */
+      stream->id = 0xBD;
+      stream->stream_type = TSMUX_ST_PRIVATE_DATA;
+      stream->pi.flags |=
+          TSMUX_PACKET_FLAG_PES_FULL_HEADER |
+          TSMUX_PACKET_FLAG_PES_DATA_ALIGNMENT;
+      break;
+    case TSMUX_ST_PS_ID3:
+      stream->id = 0xBD;        // private stream
+      stream->stream_type = TSMUX_ST_PES_METADATA;
+      stream->pi.flags |= TSMUX_PACKET_FLAG_PES_FULL_HEADER |
+          TSMUX_PACKET_FLAG_PES_DATA_ALIGNMENT;
+      break;
     case TSMUX_ST_PS_OPUS:
       /* FIXME: assign sequential extended IDs? */
       stream->id = 0xBD;
-      stream->is_audio = TRUE;
+      stream->gst_stream_type = GST_STREAM_TYPE_AUDIO;
+      stream->stream_type = TSMUX_ST_PRIVATE_DATA;
+      stream->pi.flags |= TSMUX_PACKET_FLAG_PES_FULL_HEADER;
+      break;
+    case TSMUX_ST_PS_S302M:
+      /* FIXME: assign sequential extended IDs? */
+      stream->id = 0xBD;
+      stream->gst_stream_type = GST_STREAM_TYPE_AUDIO;
+      stream->stream_type = TSMUX_ST_PRIVATE_DATA;
+      stream->pi.flags |= TSMUX_PACKET_FLAG_PES_FULL_HEADER;
+      break;
+    case TSMUX_ST_PS_VIDEO_AV1:
+      /* FIXME: assign sequential extended IDs? */
+      stream->id = 0xBD;
+      stream->gst_stream_type = GST_STREAM_TYPE_VIDEO;
       stream->stream_type = TSMUX_ST_PRIVATE_DATA;
-      stream->is_opus = TRUE;
       stream->pi.flags |= TSMUX_PACKET_FLAG_PES_FULL_HEADER;
       break;
     default:
@@ -290,6 +335,8 @@ tsmux_stream_free (TsMuxStream * stream)
   }
   g_list_free (stream->buffers);
 
+  if (stream->pmt_descriptor)
+    gst_mpegts_descriptor_free (stream->pmt_descriptor);
   g_free (stream);
 }
 
@@ -485,7 +532,7 @@ tsmux_stream_initialize_pes_packet (TsMuxStream * stream)
     }
   }
 
-  if (stream->is_video_stream) {
+  if (stream->gst_stream_type == GST_STREAM_TYPE_VIDEO) {
     guint8 hdr_len;
 
     hdr_len = tsmux_stream_pes_header_length (stream);
@@ -494,6 +541,10 @@ tsmux_stream_initialize_pes_packet (TsMuxStream * stream)
     if ((stream->cur_pes_payload_size + hdr_len - 6) > G_MAXUINT16)
       stream->cur_pes_payload_size = 0;
   }
+  // stream_type specific flags
+  if (stream->stream_type == TSMUX_ST_PES_METADATA) {
+    stream->pi.flags |= TSMUX_PACKET_FLAG_PES_DATA_ALIGNMENT;
+  }
 
   return TRUE;
 }
@@ -718,6 +769,79 @@ tsmux_stream_write_pes_header (TsMuxStream * stream, guint8 * data)
   }
 }
 
+static GstMpegtsDescriptor *
+j2k_descriptor (TsMuxStream * stream)
+{
+  GstMpegtsDescriptor *descriptor;
+  /* J2K video descriptor
+   * descriptor_tag             8 uimsbf
+   * descriptor_length          8 uimsbf
+   * profile_and_level         16 uimsbf
+   * horizontal_size           32 uimsbf
+   * vertical_size             32 uimsbf
+   * max_bit_rate              32 uimsbf
+   * max_buffer_size           32 uimsbf
+   * DEN_frame_rate            16 uimsbf
+   * NUM_frame_rate            16 uimsbf
+   * color_specification        8 bslbf
+   * still_mode                 1 bslbf
+   * interlace_video            1 bslbf
+   * reserved                   6 bslbf
+   * private_data_byte          8 bslbf
+   */
+  gint8 still_interlace_reserved = 0x00;
+  int wr_size = 0;
+  guint8 *add_info = NULL;
+  guint8 level = stream->profile_and_level & 0xF;
+  guint32 max_buffer_size = 0;
+  GstByteWriter writer;
+  gst_byte_writer_init_with_size (&writer, 32, FALSE);
+
+  switch (level) {
+    case 1:
+    case 2:
+    case 3:
+      max_buffer_size = 1250000;
+      break;
+    case 4:
+      max_buffer_size = 2500000;
+      break;
+    case 5:
+      max_buffer_size = 5000000;
+      break;
+    case 6:
+      max_buffer_size = 10000000;
+      break;
+    default:
+      break;
+  }
+
+  gst_byte_writer_put_uint16_be (&writer, stream->profile_and_level);
+  gst_byte_writer_put_uint32_be (&writer, stream->horizontal_size);
+  gst_byte_writer_put_uint32_be (&writer, stream->vertical_size);
+  gst_byte_writer_put_uint32_be (&writer, max_buffer_size);
+  gst_byte_writer_put_uint32_be (&writer, stream->max_bitrate);
+  gst_byte_writer_put_uint16_be (&writer, stream->den);
+  gst_byte_writer_put_uint16_be (&writer, stream->num);
+  gst_byte_writer_put_uint8 (&writer, stream->color_spec);
+
+  if (stream->interlace_mode)
+    still_interlace_reserved |= 0x40;
+
+  gst_byte_writer_put_uint8 (&writer, still_interlace_reserved);
+  gst_byte_writer_put_uint8 (&writer, 0x00);    /* private data byte */
+
+  wr_size = gst_byte_writer_get_size (&writer);
+  add_info = gst_byte_writer_reset_and_get_data (&writer);
+
+  descriptor =
+      gst_mpegts_descriptor_from_custom (GST_MTS_DESC_J2K_VIDEO, add_info,
+      wr_size);
+  g_free (add_info);
+
+  return descriptor;
+}
+
 /**
  * tsmux_stream_add_data:
  * @stream: a #TsMuxStream
@@ -781,7 +905,8 @@ tsmux_stream_default_get_es_descrs (TsMuxStream * stream,
   g_return_if_fail (stream != NULL);
   g_return_if_fail (pmt_stream != NULL);
 
-  if (stream->is_audio && stream->language[0] != '\0') {
+  if (stream->gst_stream_type == GST_STREAM_TYPE_AUDIO
+      && stream->language[0] != '\0') {
     descriptor = gst_mpegts_descriptor_from_iso_639_language (stream->language);
     g_ptr_array_add (pmt_stream->descriptors, descriptor);
     descriptor = NULL;
@@ -814,73 +939,16 @@ tsmux_stream_default_get_es_descrs (TsMuxStream * stream,
       break;
     case TSMUX_ST_VIDEO_JP2K:
     {
-      /* J2K video descriptor
-       * descriptor_tag             8 uimsbf
-       * descriptor_length          8 uimsbf
-       * profile_and_level         16 uimsbf
-       * horizontal_size           32 uimsbf
-       * vertical_size             32 uimsbf
-       * max_bit_rate              32 uimsbf
-       * max_buffer_size           32 uimsbf
-       * DEN_frame_rate            16 uimsbf
-       * NUM_frame_rate            16 uimsbf
-       * color_specification        8 bslbf
-       * still_mode                 1 bslbf
-       * interlace_video            1 bslbf
-       * reserved                   6 bslbf
-       * private_data_byte          8 bslbf
-       */
-      gint8 still_interlace_reserved = 0x00;
-      int wr_size = 0;
-      guint8 *add_info = NULL;
-      guint8 level = stream->profile_and_level & 0xF;
-      guint32 max_buffer_size = 0;
-      GstByteWriter writer;
-      gst_byte_writer_init_with_size (&writer, 32, FALSE);
-
-      switch (level) {
-        case 1:
-        case 2:
-        case 3:
-          max_buffer_size = 1250000;
-          break;
-        case 4:
-          max_buffer_size = 2500000;
-          break;
-        case 5:
-          max_buffer_size = 5000000;
-          break;
-        case 6:
-          max_buffer_size = 10000000;
-          break;
-        default:
-          break;
-      }
-
-      gst_byte_writer_put_uint16_be (&writer, stream->profile_and_level);
-      gst_byte_writer_put_uint32_be (&writer, stream->horizontal_size);
-      gst_byte_writer_put_uint32_be (&writer, stream->vertical_size);
-      gst_byte_writer_put_uint32_be (&writer, max_buffer_size);
-      gst_byte_writer_put_uint32_be (&writer, stream->max_bitrate);
-      gst_byte_writer_put_uint16_be (&writer, stream->den);
-      gst_byte_writer_put_uint16_be (&writer, stream->num);
-      gst_byte_writer_put_uint8 (&writer, stream->color_spec);
-
-      if (stream->interlace_mode)
-        still_interlace_reserved |= 0x40;
-
-      gst_byte_writer_put_uint8 (&writer, still_interlace_reserved);
-      gst_byte_writer_put_uint8 (&writer, 0x00);        /* private data byte */
-
-      wr_size = gst_byte_writer_get_size (&writer);
-      add_info = gst_byte_writer_reset_and_get_data (&writer);
-
-      descriptor =
-          gst_mpegts_descriptor_from_custom (GST_MTS_DESC_J2K_VIDEO, add_info,
-          wr_size);
+      descriptor = j2k_descriptor (stream);
       g_ptr_array_add (pmt_stream->descriptors, descriptor);
     }
       break;
+    case TSMUX_ST_VIDEO_JPEG_XS:
+    {
+      g_ptr_array_add (pmt_stream->descriptors,
+          gst_mpegts_descriptor_copy (stream->pmt_descriptor));
+    }
+      break;
     case TSMUX_ST_PS_AUDIO_AC3:
     {
       /* This is only called for DVB, ATSC ignores this case in favour of its
@@ -902,13 +970,33 @@ tsmux_stream_default_get_es_descrs (TsMuxStream * stream,
     case TSMUX_ST_PS_AUDIO_LPCM:
       /* FIXME */
       break;
-    case TSMUX_ST_PS_TELETEXT:
-      /* FIXME empty descriptor for now;
-       * should be provided by upstream in event or so ? */
-      descriptor =
-          gst_mpegts_descriptor_from_custom (GST_MTS_DESC_DVB_TELETEXT, 0, 1);
+    case TSMUX_ST_PES_METADATA:
 
-      g_ptr_array_add (pmt_stream->descriptors, descriptor);
+      if (stream->internal_stream_type == TSMUX_ST_PS_ID3) {
+        // metadata_descriptor
+        GstMpegtsMetadataDescriptor metadata_descriptor;
+
+        metadata_descriptor.metadata_application_format =
+            GST_MPEGTS_METADATA_APPLICATION_FORMAT_IDENTIFIER_FIELD;
+        metadata_descriptor.metadata_format =
+            GST_MPEGTS_METADATA_FORMAT_IDENTIFIER_FIELD;
+        metadata_descriptor.metadata_format_identifier = DRF_ID_ID3;
+        metadata_descriptor.metadata_service_id = 0;
+        metadata_descriptor.decoder_config_flags = 0x00;
+        metadata_descriptor.dsm_cc_flag = FALSE;
+
+        descriptor = gst_mpegts_descriptor_from_metadata (&metadata_descriptor);
+        g_ptr_array_add (pmt_stream->descriptors, descriptor);
+
+        // registration_descriptor
+        guint32 format_identifier = 0;
+        GST_WRITE_UINT32_BE (&format_identifier,
+            metadata_descriptor.metadata_format_identifier);
+
+        descriptor = gst_mpegts_descriptor_from_registration ((const char *)
+            &format_identifier, NULL, 0);
+        g_ptr_array_add (pmt_stream->descriptors, descriptor);
+      }
       break;
     case TSMUX_ST_PS_DVB_SUBPICTURE:
       /* fallthrough ...
@@ -928,7 +1016,7 @@ tsmux_stream_default_get_es_descrs (TsMuxStream * stream,
         g_ptr_array_add (pmt_stream->descriptors, descriptor);
         break;
       }
-      if (stream->is_opus) {
+      if (stream->internal_stream_type == TSMUX_ST_PS_OPUS) {
         descriptor = gst_mpegts_descriptor_from_registration ("Opus", NULL, 0);
         g_ptr_array_add (pmt_stream->descriptors, descriptor);
 
@@ -939,11 +1027,44 @@ tsmux_stream_default_get_es_descrs (TsMuxStream * stream,
 
         g_ptr_array_add (pmt_stream->descriptors, descriptor);
       }
-      if (stream->is_meta) {
+      if (stream->internal_stream_type == TSMUX_ST_PS_KLV) {
         descriptor = gst_mpegts_descriptor_from_registration ("KLVA", NULL, 0);
         GST_DEBUG ("adding KLVA registration descriptor");
         g_ptr_array_add (pmt_stream->descriptors, descriptor);
       }
+      if (stream->internal_stream_type == TSMUX_ST_PS_ST_2038) {
+        descriptor = gst_mpegts_descriptor_from_registration ("VANC", NULL, 0);
+        GST_DEBUG ("adding VANC registration descriptor");
+        g_ptr_array_add (pmt_stream->descriptors, descriptor);
+        descriptor = gst_mpegts_descriptor_from_custom (0xc4, NULL, 0);
+        g_ptr_array_add (pmt_stream->descriptors, descriptor);
+      }
+      if (stream->internal_stream_type == TSMUX_ST_PS_S302M) {
+        descriptor = gst_mpegts_descriptor_from_registration ("BSSD", NULL, 0);
+        GST_DEBUG ("adding SMPTE 302M registration descriptor");
+        g_ptr_array_add (pmt_stream->descriptors, descriptor);
+      }
+      if (stream->internal_stream_type == TSMUX_ST_PS_VP9) {
+        descriptor = gst_mpegts_descriptor_from_registration ("VP09", NULL, 0);
+        GST_DEBUG ("adding VP09 registration descriptor");
+        g_ptr_array_add (pmt_stream->descriptors, descriptor);
+      }
+      if (stream->internal_stream_type == TSMUX_ST_PS_VIDEO_AV1) {
+        /* AV1G is our custom non-bytestream format ! */
+        descriptor = gst_mpegts_descriptor_from_registration ("AV1G", NULL, 0);
+        g_ptr_array_add (pmt_stream->descriptors, descriptor);
+        if (stream->pmt_descriptor)
+          g_ptr_array_add (pmt_stream->descriptors,
+              gst_mpegts_descriptor_copy (stream->pmt_descriptor));
+      }
+      if (stream->internal_stream_type == TSMUX_ST_PS_TELETEXT) {
+        // FIXME empty descriptor for now;
+        // How should this be signalled from upstream?
+        // Definitions in  EN300468 - 6.2.43 Teletext descriptor
+        descriptor =
+            gst_mpegts_descriptor_from_custom (GST_MTS_DESC_DVB_TELETEXT, 0, 1);
+        g_ptr_array_add (pmt_stream->descriptors, descriptor);
+      }
     default:
       break;
   }
